import math
import random

import regex_ranker
import FFX

DEBUG = False

DFA_RANKING = 2
NFA_RANKING = 4


class OutputDomainTooSmallException(Exception):
    pass


class CycleWalk(Exception):
    pass


class EncryptionFailed(Exception):
    pass


def new(K, input_format, output_format, input_strategy=DFA_RANKING, output_strategy=DFA_RANKING):
    return FTE(K, input_format, output_format, input_strategy, output_strategy)


class FTE(object):

    def __init__(self, K, input_format, output_format, input_strategy, output_strategy):
        self._radix = 2

        self._ffx = FFX.new(K, radix=self._radix)

        self._K = K
        self._input_format = input_format
        self._output_format = output_format

        if output_strategy in [NFA_RANKING]:  # NFA ranking
            self._ambiguityInOutputFormat = True
        else:
            self._ambiguityInOutputFormat = False

        self._input_ranker = regex_ranker.regex_ranker(
            input_strategy, input_format[0], input_format[1])
        if input_format == output_format:
            self._fpe = True
            self._output_ranker = self._input_ranker
        else:
            self._fpe = False
            self._output_ranker = regex_ranker.regex_ranker(
                output_strategy, output_format[0], output_format[1])

        self._input_capacity = int(
            math.ceil(math.log(self._input_ranker.getNumWordsInLanguage(), self._radix)))
        self._output_capacity = int(
            math.ceil(math.log(self._output_ranker.getNumWordsInLanguage(), self._radix)))

        if self._input_ranker.getNumWordsInLanguage() > self._output_ranker.getNumWordsInLanguage():
            raise OutputDomainTooSmallException()

        self._memoization_mem = (self._input_ranker.getNumStates() * self._input_ranker.getNumSymbols() * 4)
        self._memoization_mem += (self._input_ranker.getNumStates() * int(input_format[1]) * (self._input_capacity / 8.0))
        if input_format != output_format:
            self._memoization_mem += (self._output_ranker.getNumStates() * self._output_ranker.getNumSymbols() * 4)
            self._memoization_mem += (self._output_ranker.getNumStates() * int(output_format[1]) * (self._output_capacity / 8.0))
        
        self._outputrangeerror = 0
        self._rankfail = 0

    def encrypt(self, plaintext, tweak=None):
        if DEBUG:
            print ['encrypt.message', plaintext]

        if tweak == None:
            tweak = FFX.FFXInteger('0' * 128)

        key = FFX.FFXInteger(self._K, radix=2, blocksize=128)

        padding_ctr = 0
        message_rank = self._input_ranker.rank(plaintext)
        #while True:
        if self._output_capacity > self._input_capacity:
            padding = FFX.FFXInteger(
                padding_ctr, radix=self._radix, blocksize=self._output_capacity - self._input_capacity)
            message = FFX.FFXInteger(
                message_rank, radix=self._radix, blocksize=self._input_capacity)
            message = FFX.FFXInteger(
                padding.to_str() + message.to_str(), radix=self._radix, blocksize=self._output_capacity)
        else:
            message = FFX.FFXInteger(
                message_rank, radix=self._radix, blocksize=self._output_capacity)

        retval = self._ffx.encrypt(tweak, message)
        while True:
            try:
                # test using simple comparison to determine if FFX.encrypt
                # has valid value for unranking
                if retval.to_int() >= self._output_ranker.getNumWordsInLanguage():
                    self._outputrangeerror += 1
                    raise CycleWalk()

                _tmp = self._output_ranker.unrank(retval.to_int())

                if self._ambiguityInOutputFormat:
                    if retval.to_int() != self._output_ranker.rank(_tmp):
                        if retval.to_int() < self._input_ranker.getNumWordsInLanguage():
                            _tmp2 = self._input_ranker.unrank(retval.to_int())
                            badPoint = (retval.to_int() == self._input_ranker.rank(_tmp2))
                        else:
                            badPoint = False
                        if badPoint:
                            raise EncryptionFailed()
                        else:
                            self._rankfail += 1
                            raise CycleWalk()

                retval = _tmp
                break
            except CycleWalk:
                retval = self._ffx.encrypt(tweak, retval)

        if DEBUG:
            print ['encrypt.retval', retval]

        return retval

    def decrypt(self, ciphertext, tweak=None):
        if DEBUG:
            print ['decrypt.ciphertext', ciphertext]

        if tweak == None:
            tweak = FFX.FFXInteger('0' * 128)

        key = FFX.FFXInteger(self._K, radix=2, blocksize=128)

        ciphertext = self._output_ranker.rank(ciphertext)
        ciphertext = FFX.FFXInteger(
            ciphertext, radix=self._radix, blocksize=self._output_capacity)
        while True:
            try:
                ciphertext = self._ffx.decrypt(tweak, ciphertext)

                if ciphertext.to_int() >= self._input_ranker.getNumWordsInLanguage():
                    raise CycleWalk()

                _tmp = self._input_ranker.unrank(ciphertext.to_int())
                if ciphertext.to_int() != self._input_ranker.rank(_tmp):
                    raise CycleWalk()

                retval = self._input_ranker.unrank(ciphertext[-1 * self._input_capacity:].to_int())
                break
            except CycleWalk:
                continue

        if DEBUG:
            print ['decrypt.retval', retval]

        return retval
