#include <chrono>
#include <gmp.h>
#include <gmpxx.h>
#include "performance_benchmark_math.h"
#include "performance_benchmark_mem.h"
#include "performance_benchmark_timer.h"
#include "regex_ranker.h"

inline size_t englishToStrategy(std::string strategy) {
    size_t retval;
    
    if(strategy=="dfa_ranking") {
        retval = regex_ranker_strategy::DFA_RANKING2; 
    } else if(strategy=="nfa_ranking") {
        retval = regex_ranker_strategy::NFA_RANKING; 
    }
    
    return retval;
}

inline std::string strategyToEnglish(size_t strategy) {
    std::string retval = "";
    switch (strategy) {
    case regex_ranker_strategy::DFA_RANKING2:
        retval = "dfa_ranking";
        break;
    case regex_ranker_strategy::NFA_RANKING:
        retval = "nfa_ranking";
        break;
    }
    return retval;
}

inline std::string str_replace( const std::string X,
                                const char find,
                                const std::string replace ) {
    std::string retval = "";
    for (size_t i = 0; i<X.size(); i++) {
        if (X[i]==find) {
            retval += replace;
        } else {
            retval += X[i];
        }
    }
    return retval;
}

inline void log_trial( size_t strategy,
                       std::string lanuage_descriptor,
                       size_t max_len,
                       std::string function,
                       size_t trials,
                       std::vector<double> perf_results,
                       std::vector<uint32_t> mem_results ) {

    std::cout << strategyToEnglish(strategy) << ",";
    std::cout << "\"" + str_replace(lanuage_descriptor,' ',"\\s") + "\"" << ",";
    std::cout << max_len << ",";
    std::cout << function << ",";
    std::cout << trials << ",";
    std::cout << mean(perf_results) << ",";
    std::cout << median(perf_results) << ",";
    std::cout << min(perf_results) << ",";
    std::cout << max(perf_results) << ",";
    std::cout << mean(mem_results) << std::endl;
}

void doInitTest( size_t strategy, std::string regex, size_t max_len, size_t trials ) {
    size_t i = 0;
    std::vector<double> perf_results;
    std::vector<uint32_t> mem_results;
    for (i = 0; i < trials; i++) {
        performance_benchmark_mem m;
        performance_benchmark_timer t;
        regex_ranker * rankerObj = new regex_ranker(strategy, regex, max_len);
        perf_results.push_back( t.elapsed() );
        mem_results.push_back( m.consumed() );
        delete rankerObj;
    }
    log_trial( strategy, regex, max_len, "memoization", trials, perf_results, mem_results );
}

void doUnrankTest( size_t strategy, std::string regex, size_t max_len, size_t trials ) {

    rankerobject * rankerObj = new regex_ranker(strategy, regex, max_len);

    size_t i = 0;
    std::vector<double> perf_results;
    std::vector<uint32_t> mem_results;
    for (i = 0; i < trials; i++) {
        mpz_class n = rankerObj->getNumWordsInLanguage();
        mpz_class C = random_integer(n);

        performance_benchmark_mem m;
        performance_benchmark_timer t;
        std::string X = rankerObj->unrank(C);
        perf_results.push_back( t.elapsed() );
        mem_results.push_back( m.consumed() );

        mpz_class D = rankerObj->rank(X);
        assert(D == C);
    }

    delete rankerObj;

    log_trial( strategy, regex, max_len, "unrank", trials, perf_results, mem_results );
}

void doRankTest( size_t strategy, std::string regex, size_t max_len, size_t trials ) {

    rankerobject * rankerObj = new regex_ranker(strategy, regex, max_len);

    size_t i = 0;
    std::vector<double> perf_results;
    std::vector<uint32_t> mem_results;
    for (i = 0; i < trials; i++) {
        mpz_class n = rankerObj->getNumWordsInLanguage();
        mpz_class C = random_integer(n);
        std::string X = rankerObj->unrank(C);

        performance_benchmark_mem m;
        performance_benchmark_timer t;
        mpz_class D = rankerObj->rank(X);
        perf_results.push_back( t.elapsed() );
        mem_results.push_back( m.consumed() );

        assert(D == C);
    }

    delete rankerObj;

    log_trial( strategy, regex, max_len, "rank", trials, perf_results, mem_results );
}

void help() {
    std::cout << "usage: performance_benchmark [strategy] [algorithm] [regex] [max_len]" << std::endl;
    std::cout << "  strategy = dfa_ranking | nfa_ranking" << std::endl;
    std::cout << "  algorithm = memoization | rank | unrank" << std::endl;
    std::cout << "  regex = [user defined regular expression R]" << std::endl;
    std::cout << "  max_len = maximum string length in L(R) to compute for memoization" << std::endl;
    exit(1);
}

int main(int argc, char** argv) {
    if (argc!=5) {
        help();
    }
    
    size_t strategy = englishToStrategy(argv[1]);
    std::string algorithm = argv[2];
    std::string regex = argv[3];
    size_t max_len = atoi(argv[4]);

    size_t num_trials_init = 8;
    size_t num_trials_rank = 8192;
    
    if (algorithm=="memoization") {
        doInitTest( strategy, regex, max_len, num_trials_init );
    } else if (algorithm=="unrank") {
        doUnrankTest( strategy, regex, max_len, num_trials_rank );
    } else if (algorithm=="rank") {
        doRankTest( strategy, regex, max_len, num_trials_rank );
    }

    return 0;
}
