#ifdef PYTHON_MODULE
#include <boost/python/module.hpp>
#include <boost/python/def.hpp>
#include <boost/python/class.hpp>
#endif

#include "regex_ranker.h"

#include <gmpxx.h>

static class _strategy_unknown_exception: public std::exception
{
    virtual const char* what() const throw()
    {
        return "Invalid regex strategy specified.";
    }
} strategy_unknown_exception;

regex_ranker::regex_ranker(const uint32_t strategy,
                           const std::string regex,
                           const uint32_t max_len) :
    _strategy(strategy),
    _regex(regex),
    _max_len(max_len),
    _num_states(0)
{
    switch (strategy) {
    case DFA_RANKING1:
    case DFA_RANKING2:
    case NFA_RANKING:
        _regex = "/" + regex + "/";
        break;
    }
    
    switch (strategy) {
    case DFA_RANKING1:
        _ranker = new dfa_rank(_regex, _max_len);
        break;
    case DFA_RANKING2:
        _ranker = new ffa_rank(_regex, _max_len);
        break;
    case NFA_RANKING:
        _ranker = new nfa_rank(_regex, _max_len);
        break;
    }
    _num_states = _ranker->_num_states;
    _num_symbols = 256;
}

regex_ranker::~regex_ranker() {
    delete _ranker;
}

#ifdef PYTHON_MODULE
std::string regex_ranker::unrank(PyObject* C) {
    std::string retval = "";
    
    // PyNumber to mpz_class
    int base = 16;
    PyObject* b16 = PyNumber_ToBase(C, base);
    const char* the_c_str = PyString_AsString(b16);
    mpz_class to_unrank(the_c_str, 0);
    
    // Rank mpz_class
    retval = _ranker->unrank(to_unrank);
    
    return retval;
}

PyObject* regex_ranker::rank(const std::string X) {
    PyObject* retval;
    
    mpz_class n = _ranker->rank(X);
    
    // Set our c_out value
    uint32_t base = 10;
    std::string to_convert = n.get_str(base);
    retval = PyLong_FromString((char*)(to_convert.c_str()), NULL, base);
    
    return retval;
}

PyObject* regex_ranker::getNumWordsInLanguage() {
    PyObject* retval;
    
    mpz_class n = _ranker->getNumWordsInLanguage();
    
    // Set our c_out value
    uint32_t base = 10;
    std::string to_convert = n.get_str(base);
    retval = PyLong_FromString((char*)(to_convert.c_str()), NULL, base);
    
    return retval;
}

PyObject* regex_ranker::getT(unsigned int state, size_t i) {
    PyObject* retval;
    
    mpz_class n = _ranker->tab[state][i];
    
    // Set our c_out value
    uint32_t base = 10;
    std::string to_convert = n.get_str(base);
    retval = PyLong_FromString((char*)(to_convert.c_str()), NULL, base);
    
    return retval;
}
#else
std::string regex_ranker::unrank(const mpz_class C) {
    return _ranker->unrank(C);
}

mpz_class regex_ranker::rank(const std::string X) {
    return _ranker->rank(X);
}

mpz_class regex_ranker::getNumWordsInLanguage() {
    return _ranker->getNumWordsInLanguage();
}

mpz_class regex_ranker::getT(unsigned int state, size_t i) {
    return _ranker->tab[state][i];
}
#endif

uint32_t regex_ranker::getNumStates() {
    return _num_states;
}

uint32_t regex_ranker::getNumSymbols() {
    return _num_symbols;
}

#ifdef PYTHON_MODULE
BOOST_PYTHON_MODULE(regex_ranker)
{
    boost::python::class_<regex_ranker>("regex_ranker",boost::python::init<uint32_t,std::string,uint32_t>())
    .def("rank", &regex_ranker::rank)
    .def("unrank", &regex_ranker::unrank)
    .def("getNumStates", &regex_ranker::getNumStates)
    .def("getNumSymbols", &regex_ranker::getNumSymbols)
    .def("getNumWordsInLanguage", &regex_ranker::getNumWordsInLanguage)
    .def("getT", &regex_ranker::getT);
}
#endif
